
(*: Name : PolygonTessellation` *)

(*: Title : PolygonTessellation *)

(*: Author : Martin Kraus *)

(*: Package Version : 1.2 *)

(*: Mathematica Version : 4.0 *)

(*: Summary : 
      "This package provides functions to tessellate polygons in two and \
three dimensions. The tessellated polygons may then be triangulated with \
functions from the package SimplePolygonTriangulation;though,this might not \
work in all cases." *)

(*: History : 
      "Created August 1999 by Martin Kraus (mk, \
mailto:Martin_Kraus_Germany@yahoo.com), 
09/06/99 mk version 1.1: cleaned up examples and replaced $MachineEpsilon by \
Chop
01/07/99 mk version 1.2: new context PolygonTriangulation` "*)

(*: Keywords : "geometry,polygon,tessellation" *)

(*: Source : 
      "See source in 'SimplePolygonTriangulation.m' and some ideas from \
Woo,Neider,Davis,'OpenGL Programming Guide',2nd ed.,chapter 11" *)

(*: Warnings : See Limitation. *)

(*: Limitation : 
      "Only one (possibly self-intersecting) polygon can be tessellated.There \
are (exotic) polygons which are not tessellated into perfectly \
non-self-intersecting polygons." *)

(*: Discussion :
      "This (very simple) implementation builds up a directed graph of \
non-identical vertices with non-crossing edges (however,multiple edges \
between two vertices are allowed).In a second step a tessellation is created \
according to a user-defined winding number rule.Future implementations should \
be able to tessellate several unconnected polygons, like the tessellation in \
the GLU library of OpenGL 1.2" *)

BeginPackage["PolygonTriangulation`PolygonTessellation`", 
    "PolygonTriangulation`SimplePolygonTriangulation`"];

(* Usage Messages *)

PolygonTessellation::usage = 
    "PolygonTessellation[pts] returns a tessellation of the (possibly \
self-intersecting) polygon specified by the list of two-dimensional points \
pts. The tessellation consists of of a list of lists of two-dimensional \
points specifying the resulting polygons which have no self-intersections.";

PolygonTessellation3D::usage  = 
    "PolygonTessellation3D[pts] returns a tessellation of the (possibly \
self-intersecting) polygon specified by the list of three-dimensional points \
pts. The tessellation consists of a list of lists of two-dimensional points \
specifying the resulting polygons which have no self-intersections.";

TessellatePolygons::usage = 
    "TessellatePolygons[expr] returns expr with tessellated polygons assuming \
all polygons in expr are two-dimensional.";

TessellatePolygons3D::usage = 
    "TessellatePolygons3D[expr] returns expr with tessellated polygons \
assuming all polygons in expr are three-dimensional and flat (see also option \
FlatPolygons).";

WindingRule::usage = 
    "WindingRule is an option for tessellating functions which specifies a \
function to apply to winding numbers to determine whether the corresponding \
polygons should be generated. The default setting WindingRule->OddQ will \
generate polygons with an odd winding number only. Other common settings are \
Positive, Negative, (# != 0 &),  (Abs[#] >= 2 &), EvenQ, (True &)";

FlatPolygons::usage = 
    "FlatPolygons is an option for TessellatePolygons3D. FlatPolygons->All \
will tessellate all polygons. FlatPolygons->Automatic will only tessellate \
Polygon primitives with a second argument specifying the normal vector of the \
polygon (this second argument may be Automatic).";

(* Options *)

Options[PolygonTessellation] = {WindingRule -> OddQ, 
      GraphicalDebugging -> False};

Options[PolygonTessellation3D] = {PlaneNormal -> Automatic};

Options[TessellatePolygons] = {};

Options[TessellatePolygons3D] = {WindingRule -> OddQ, 
      GraphicalDebugging -> False, FlatPolygons -> All};
(* Error Messages *)

PolygonTessellation::internal = "An internal error (code `1`) occured.";

Begin["`Private`"];

Off[General::spell1];
(* Auiliary Functions *)

(* "order: point, node, edge, graph, tessellation, graphics" *)

(* import functions from SimplePolygonTriangulation *)

normalize = 
    PolygonTriangulation`SimplePolygonTriangulation`Private`normalize;
getPlane = PolygonTriangulation`SimplePolygonTriangulation`Private`getPlane;
embedPointIn3D = 
    PolygonTriangulation`SimplePolygonTriangulation`Private`embedPointIn3D;
projectPointToPlane = 
    PolygonTriangulation`SimplePolygonTriangulation`Private`\
projectPointToPlane;

(* "returns True iff {x,y} is on one line with nodes n1 and n2" *)

pointCollinearQ[{x_, y_}, n1_, n2_] := 
    Module[{dx1, dy1, dx2, dy2}, dx1 = x - nodes[[n1, 1, 1]]; 
      dy1 = y - nodes[[n1, 1, 2]]; dx2 = nodes[[n2, 1, 1]] - x; 
      dy2 = nodes[[n2, 1, 2]] - y; 
      If[Chop[Abs[dx1*dy2 - dy1*dx2]] <= 0, True, False]];

(* "returns True iff {x,y} is on the edge between nodes n1 and n2" *)

pointOnEdgeQ[{x_, y_}, n1_, n2_] := 
    Module[{}, If[! pointCollinearQ[{x, y}, n1, n2], Return[False]]; 
      If[(Chop[x - nodes[[n1, 1, 1]]] >= 0 || 
              Chop[x - nodes[[n2, 1, 1]]] >= 
                0) && (Chop[x - nodes[[n1, 1, 1]]] <= 0 || 
              Chop[x - nodes[[n2, 1, 1]]] <= 
                0) && (Chop[y - nodes[[n1, 1, 2]]] >= 0 || 
              Chop[y - nodes[[n2, 1, 2]]] >= 
                0) && (Chop[y - nodes[[n1, 1, 2]]] <= 0 || 
              Chop[y - nodes[[n2, 1, 2]]] <= 0), True, False]];
(* "nodeAllEdges returns a list of all directed edges as a list pairs of node \
indices" *)

nodeAllEdges[] := 
    Flatten[MapIndexed[
        Transpose[{Table[#2[[1]], {Length[#1[[2]]]}], #1[[2]]}] &, nodes], 
      1];

(* "nodeInsert inserts a new vertex into the list of nodes, returning the \
index of the new node in nodes." *)

nodeInsert[{x_, y_}] := 
    Module[{occurrences, allEdges, n1, n2, nx, ny, i, outEdges, px, 
        py}, {nx, ny} = N[{x, y}]; (* 
        look for an existing vertex with this coordinates *)  
      occurrences = 
        Position[nodes, 
          node[{px_, py_}, ___] /; Chop[{px - nx, py - ny}] == {0, 0}]; 
      If[Length[occurrences] > 1, Throw["msgins1", PolygonTessellation]]; 
      If[Length[occurrences] == 1, Return[First[First[occurrences]]]]; 
      allEdges = nodeAllEdges[]; 
      occurrences = 
        Extract[allEdges, 
          Position[Map[pointOnEdgeQ[{nx, ny}, #[[1]], #[[2]]] &, allEdges], 
            True]];
      If[Length[occurrences] > 0,(* the {nx, ny} is on an edge *) 
        inEdges = {}; outEdges = {}; 
        For[i = 1, i <= Length[occurrences], i++,
          {n1, n2} = 
            occurrences[[i]];(* {nx, ny} is on an edge from n1 to n2 *)
      
              nodes[[n1]] = 
            node[nodes[[n1, 1]], nodes[[n1, 2]] /. n2 :> numberOfNodes + 1];
          AppendTo[outEdges, n2]]; AppendTo[nodes, node[{nx, ny}, outEdges]]; 
        numberOfNodes = numberOfNodes + 1; Return[numberOfNodes]];(* 
        new point is not identical to a vertex nor on an edge *) 
      AppendTo[nodes, node[{nx, ny}, {}]]; numberOfNodes = numberOfNodes + 1; 
      numberOfNodes];

(* "returns True if node n1 is left of n2" *)

nodeLeftOfQ[n1_, n2_] := OrderedQ[{nodes[[n1, 1]], nodes[[n2, 1]]}];

(* "sets the winding number of a specified edge and all edges to the same \
node assuming that the first edge was given. Also, the corresponding winding \
numbers in the other node are set. This function takes care of all the \
problems associated with multiple edges between two nodes. The function \
returns the node and edge indices of the edge of the other node which \
encloses the same polygon as the specified edge." *)

nodeSetWindingNumber[node_, edge_, windingNumber_] := 
    Module[{i, currentWN, toNode, nextNode, nextEdge}, 
      If[FreeQ[neighboringNodes, node], AppendTo[neighboringNodes, node]]; 
      toNode = Abs[nodes[[node, 3, edge]]]; i = edge; 
      currentWN = windingNumber; 
      j = First[Last[Position[nodes[[toNode, 3]], node | -node]]]; 
      If[j < Length[nodes[[toNode, 3]]], j++, j = 1]; nextNode = toNode; 
      nextEdge = j; 
      While[i <= Length[nodes[[node, 3]]] && 
          Abs[nodes[[node, 3, i]]] == toNode, 
        If[(nodes[[node, 4, i]] != Indeterminate && 
                nodes[[node, 4, i]] != currentWN) || (nodes[[toNode, 4, j]] !=
                   Indeterminate && nodes[[toNode, 4, j]] != currentWN), 
          Throw["msgswn1", PolygonTessellation]]; 
        nodes[[node, 4, i]] = currentWN; nodes[[toNode, 4, j]] = currentWN; 
        If[currentWN === Infinity, currentWN = 0]; 
        If[nodes[[node, 3, i]] > 0, currentWN--, currentWN++]; i++; 
        If[j > 1, j--, j = Length[nodes[[toNode, 3]]]]]; {nextNode, 
        nextEdge}];
(* "determines the edge index and winding number of the next unprocessed edge \
and returns them in a list or {} if all edges in this node are processed (the \
node is then dropped from the neighboringNodes list)" *)

nodeNextEdgeAndWindingNumber[node_] := 
    Module[{pos, edge, previousWN, windingNumbers, currentWN}, 
      windingNumbers = nodes[[node, 4]]; 
      If[FreeQ[windingNumbers, Indeterminate], 
        If[FreeQ[neighboringNodes, node], 
          Throw["msgnxt2", PolygonTessellation], 
          neighboringNodes = 
            Delete[neighboringNodes, Position[neighboringNodes, node]]]; 
        Return[{}]]; 
      If[FreeQ[windingNumbers, _Integer], 
        Throw["msgnxt1", PolygonTessellation]];(* 
        set pos to the last integer before Indeterminate and edge to the \
Indeterminate position *) 
      pos = First[Last[Position[windingNumbers, _Integer]]]; 
      If[pos < Length[windingNumbers], edge = pos + 1, 
        edge = First[First[Position[windingNumbers, Indeterminate]]]; 
        If[edge > 1, pos = edge - 1, pos = Length[windingNumbers]]]; 
      previousWN = windingNumbers[[pos]]; 
      If[previousWN === Infinity, previousWN = 0]; 
      If[nodes[[node, 3, pos]] > 0, currentWN = previousWN - 1, 
        currentWN = previousWN + 1]; {edge, currentWN}];

(* "inserts an edge into the directed graph taking care of existing nodes on \
the edge (does not worry about intersecting edges)" *)

edgeRecursiveInsert[n1_, n2_] := 
    Module[{i = 1}, 
      While[i <= 
            numberOfNodes && (i == n1 || 
              i == n2 || ! pointOnEdgeQ[nodes[[i, 1]], n1, n2]), 
        i++];(*Print[n1, " ", n2, " ", i, " ", numberOfNodes];*) 
      If[i > Length[nodes],(* no nodes on this edge *) 
        nodes[[n1]] = node[nodes[[n1, 1]], Append[nodes[[n1, 2]], n2]]; 
        AppendTo[newEdges, {n1, n2}], (* 
          node i is somewhere between : recurse *) edgeRecursiveInsert[n1, i];
         edgeRecursiveInsert[i, n2]]];

(* "inserts an edge into the directed graph" *)

edgeInsert[n1_, n2_] := 
    Module[{i, j, allEdges, processedEdges = {}, m1, m2, k1, k2, point, 
        outEdges}, If[n1 == n2, Return[]]; 
      newEdges = {};(* partition the edge at crossed nodes *) 
      edgeRecursiveInsert[n1, n2];(* 
        intersect the resulting parts with all other edges *) 
      allEdges = nodeAllEdges[]; 
      For[i = 1, i <= Length[allEdges], i++, {m1, m2} = allEdges[[i]]; 
        For[j = 1, j <= Length[newEdges], j++, {k1, k2} = newEdges[[j]]; 
          If[FreeQ[processedEdges, Sort[{m1, m2}]], 
            point = edgeIntersection[m1, m2, k1, k2]; 
            If[point != {},(* insert new node *)  
              AppendTo[processedEdges, Sort[{m1, m2}]]; 
              outEdges = 
                Flatten[{Cases[nodes[[m1, 2]], m2], k2, 
                    Cases[nodes[[m2, 2]], m1]}]; 
              AppendTo[nodes, node[point, outEdges]]; 
              numberOfNodes = numberOfNodes + 1; 
              nodes[[m1]] = 
                node[nodes[[m1, 1]], nodes[[m1, 2]] /. m2 :> numberOfNodes]; 
              nodes[[m2]] = 
                node[nodes[[m2, 1]], nodes[[m2, 2]] /. m1 :> numberOfNodes]; 
              nodes[[k1]] = 
                node[nodes[[k1, 1]], nodes[[k1, 2]] /. k2 :> numberOfNodes];(* 
                update newEdges *) newEdges[[j]] = {k1, numberOfNodes}; 
              AppendTo[newEdges, {numberOfNodes, k2}];(* 
                no other of the new edges can be intersected by m1, m2 *) 
              j = Length[newEdges]]]]]];

(* "returns {} or the coordinates of the intersection of the edges from node \
n1 to n2 and m1 to m2" *)
  edgeIntersection[n1_, n2_, m1_, m2_] := 
    Module[{s, t, det, ax, ay, bx, by, cx, cy, dx, dy}, 
      If[Length[Union[{n1, n2, m1, m2}]] != 4,(* duplicated index *) 
        Return[{}]]; {ax, ay} = nodes[[n1, 1]]; {bx, by} = 
        nodes[[n2, 1]]; {cx, cy} = nodes[[m1, 1]]; {dx, dy} = nodes[[m2, 1]]; 
      det = (bx - ax)*(cy - dy) - (cx - dx)*(by - ay); 
      If[Chop[Abs[det]] <= 0, Return[{}]]; 
      s = ((cx - ax)*(cy - dy) - (cy - ay)*(cx - dx))/det; 
      t = ((cy - ay)*(bx - ax) - (cx - ax)*(by - ay))/det; 
      If[Chop[s] < 0 || Chop[s - 1] > 0 || Chop[t] < 0 || Chop[t - 1] > 0, 
        Return[{}]];(*Print[s, " ", n1, nodes[[n1, 1]], "->", n2, 
            nodes[[n2, 1]], " ", m1, nodes[[m1, 1]], "->", m2, 
            nodes[[m2, 1]]];*) Return[{ax, ay} + ({bx, by} - {ax, ay})*s]];

(* "returns True if the edge from n1 to n2 is 'left' ('anti-clockwise turn') \
of the edge from n1 to n3. 'left' means that Arg of the corresponding complex \
number is greater; thus, the left-most edge is the one going to the left and \
slightly upwards; the right-most is going to the left and slightly downwards. \
Negative indices are allowed." *) 
edgeLeftOfQ[n1_, n2_, n3_] := 
    Module[{x1, y1, x2, y2}, 
      If[Abs[n1] == Abs[n2] || Abs[n1] == Abs[n3], 
        Throw["msglef1", PolygonTessellation]]; 
      If[Abs[n2] == Abs[n3], Return[If[n2 > n3, True, False]]]; {x1, y1} = 
        nodes[[Abs[n2], 1]] - nodes[[Abs[n1], 1]]; {x2, y2} = 
        nodes[[Abs[n3], 1]] - nodes[[Abs[n1], 1]]; 
      If[Arg[x1 + I y1] > Arg[x2 + I y2], True, False]];
(* "returns a node structure with the sorted edges inserted. These sorted \
edges start with the 'left-most'." *)

graphInsertSortedEdges[n_] := 
    Module[{outEdges, inEdges, sortedEdges}, outEdges = nodes[[n, 2]]; 
      inEdges = 
        Flatten[MapIndexed[Table[#2[[1]], {Length[Cases[#1[[2]], n]]}] &, 
            nodes]]; 
      If[Length[inEdges] != Length[outEdges], 
        Throw["msgsor1", PolygonTessellation]]; 
      sortedEdges = 
        Sort[Join[outEdges, -inEdges], (edgeLeftOfQ[n, #1, #2]) &]; 
      node[nodes[[n, 1]], nodes[[n, 2]], sortedEdges, 
        Table[Indeterminate, {Length[sortedEdges]}]]];

(* "sorts edges preparing the graph for tessellation. Returns the index of \
the left-most node." *)

graphSortEdgesAndNodes[] := Module[{sortedNodeIndices},
      nodes = Map[graphInsertSortedEdges, Range[numberOfNodes]]; 
      sortedNodeIndices = Sort[Range[numberOfNodes], nodeLeftOfQ]; 
      First[sortedNodeIndices]];
graphSimplePolygons[] := 
    Module[{node, next, nextEdge, nextWindingNumber, polygon}, 
      While[Length[neighboringNodes] > 0, node = First[neighboringNodes]; 
        next = nodeNextEdgeAndWindingNumber[
            node];(*Print[next, neighboringNodes];*)If[
          next != {}, {nextEdge, nextWindingNumber} = next; 
          polygon = graphCreatePolygon[node, nextEdge, nextWindingNumber]; 
          If[windingRule[nextWindingNumber] === True, 
            AppendTo[polygons, polygon]]]]; polygons];
graphCreatePolygon[startNode_, startEdge_, windingNumber_] := 
    Module[{node, edge, iteration, polygon}, polygon = {}; node = startNode; 
      edge = startEdge; iteration = 1; 
      While[(iteration == 1 || node != startNode || edge != startEdge) && 
          iteration <= 2*numberOfNodes, AppendTo[polygon, node]; 
        If[graphicalDebugging === True, 
          graphGraphics[polygon]]; {node, edge} = 
          nodeSetWindingNumber[node, edge, windingNumber]; iteration++]; 
      If[iteration > 2*numberOfNodes, Throw["msgcre", PolygonTessellation]]; 
      polygon];
graphMarkOutline[leftMostNode_] := 
    Module[{}, 
      outline = Reverse[graphCreatePolygon[leftMostNode, 1, Infinity]]];
(* "tessellate one polygon in three dimensions using the options of \
TessellatePolygons3D. edges are not yet supported." *)

tessellatePolygon3D[poly : Polygon[points_: {}, normal_: None, edges_: {}], 
      opts___] := 
    Module[{graphicalDebugging, flatPolygons, realNormal, tessellation, 
        i}, {graphicalDebugging, windingRule, 
          flatPolygons} = {GraphicalDebugging, WindingRule, 
              FlatPolygons} /. {opts} /. Options[TessellatePolygons3D]; 
      If[Length[points] <= 
            3 || (normal === None && flatPolygons === Automatic), 
        Return[poly]]; 
      If[normal === None, realNormal = Automatic, realNormal = normal]; 
      tessellation = 
        PolygonTessellation3D[points, WindingRule -> windingRule, 
          GraphicalDebugging -> graphicalDebugging, 
          PlaneNormal -> realNormal]; If[catchResult =!= True, Return[poly]]; 
      Return[Table[
          If[normal === None, Polygon[tessellation[[i]]], 
            If[edges == {} || i != 1, Polygon[tessellation[[i]], realNormal], 
              Polygon[tessellation[[i]], realNormal, edges]]], {i, 1, 
            Length[tessellation]}]]];
graphGraphics[polygon_:{}] := 
    Module[{allEdges, index, xmin, xmax, ymin, ymax, width, graphicsEdges, 
        graphicsNodes, graphicsOutline, graphicsOptions, graphicsPolygons, 
        graphicsPolygon}, xmin = Min[Transpose[rawPoints][[1]]]; 
      xmax = Max[Transpose[rawPoints][[1]]]; 
      ymin = Min[Transpose[rawPoints][[2]]]; 
      ymax = Max[Transpose[rawPoints][[2]]]; width = xmax - xmin; 
      graphicsOptions = 
        Graphics[{}, 
          PlotRange -> {{xmin - 0.05(xmax - xmin), 
                xmax + 0.05(xmax - xmin)}, {ymin - 0.05(ymax - ymin), 
                ymax + 0.05(ymax - ymin)}}, AspectRatio -> Automatic, 
          Axes -> True]; 
      graphicsNodes = 
        Graphics[{PointSize[0.03], Map[Point[#[[1]]] &, nodes]}]; 
      allEdges = nodeAllEdges[]; 
      graphicsEdges = 
        Graphics[{Thickness[0.01], 
            MapIndexed[graphGraphicsEdge[#1[[1]], #1[[2]], #2[[1]], width] &, 
              allEdges]}]; 
      graphicsOutline = 
        If[outline == {}, Graphics[{}], 
          Graphics[{RGBColor[0.8, 0.8, 0.8], 
              Polygon[Map[(nodes[[#, 1]]) &, outline]]}]]; 
      graphicsPolygons = 
        If[polygons == {}, Graphics[{}], 
          Graphics[{RGBColor[0.5, 0.3, 1], Thickness[0.01], 
              Map[Polygon, polygons /. index_Integer :> nodes[[index, 1]]]}]];
       graphicsPolygon = 
        If[polygon == {}, Graphics[{}], 
          Graphics[{RGBColor[1, 0, 0], Thickness[0.015], 
              Line[Map[(nodes[[#, 1]]) &, polygon]]}]]; 
      Show[graphicsOptions, graphicsOutline, graphicsPolygons, graphicsNodes, 
        graphicsEdges, graphicsPolygon]];

graphGraphicsEdge[n1_, n2_, edgeIndex_, width_] := 
  Module[{x1, y1, x2, y2, pos, basis, perp, length}, {x1, y1} = 
      nodes[[n1, 1]]; {x2, y2} = nodes[[n2, 1]]; perp = {-(y2 - y1), x2 - x1};
     pos = 0.5 + 0.4*FractionalPart[edgeIndex*3.14159265358979]; 
    basis = {x1, y1} + pos({x2, y2} - {x1, y1}); 
    length = Sqrt[(x2 - x1)^2 + (y2 - y1)^2]; 
    If[length == 0., 
      Throw["msgged1", PolygonTessellation]]; {Line[{{x1, y1}, {x2, y2}}], 
      Polygon[{basis + 0.05*width*({x2, y2} - {x1, y1})/length, 
          basis + 0.5*0.05*width*perp/length, 
          basis - 0.5*0.05*width* perp/length}]}];
(* Local (Static) Variables *)

(* "nodes is a list of node objects. Each node object has the form \
node[{x,y},{nodeOut1,nodeOut2,nodeOut3,...}] in the first step of building up \
the directed graph and the form \
node[{x,y},{nodeOut1,nodeOut2,nodeOut3,...},{nodeInOut1,nodeInOut2,nodeInOut3,\
...},{windingNumber1,windingNumber2,...}] where nodeOutX are indices of nodes \
of outgoing edges, nodeInOutX are sorted according to the orientation of the \
edge and negative for incoming edges; windingNumberX corrsponds to the sorted \
edges and give the winding number for polygons at the left-hand side \
(anti-clockwise) of the corresponding edge. A value unequal from \
Indeterminate indicates that the polygon has been processed already." *)

nodes = {};

(* "numberOfNodes is equal to Length[nodes]" *)
numberOfNodes = 0;

(* "rawPoints is the list of points defining the self-intersecting polygon" *) \

rawPoints = {};

(* "neighboringNodes is a list of the indices of all nodes which have been \
visited during the tessellation but not completely processed yet. If this \
list is empty, the tessellation is finished." *)
neighboringNodes = {};

(* "graphicalDebugging is True when graphics should be produced." *)
  graphicalDebugging = False;

(* "windingRule is called to decide whether a polygon of a particular winding \
number should be created." *)
windingRule = OddQ;

(* "outline is a list of node indices, which form the non-convex outline of \
the self-intersecting polygon" *) 
outline = {};

(* "list of lists of points, defining simple polygons." *)
polygons = {};

(* "error code of PolygonTessellation, True if no error occurred" *)
  catchResult = True;

(* Public Functions (see usage messages for comments) *)

PolygonTessellation[pts_, opts___] := 
    Module[{n1, n2, 
        leftMostNode}, {graphicalDebugging, 
          windingRule} = {GraphicalDebugging, WindingRule} /. {opts} /. 
          Options[PolygonTessellation]; rawPoints = N[pts]; 
      If[Last[rawPoints] != First[rawPoints], 
        AppendTo[rawPoints, First[rawPoints]]]; polygons = {}; outline = {}; 
      nodes = {}; numberOfNodes = 0; 
      catchResult = 
        Catch[n1 = nodeInsert[First[rawPoints]];(* 
            firststep : 
              building up the directed graph with non - intersecting edges *) 
          For[i = 2, i <= Length[rawPoints], i++, 
            n2 = nodeInsert[rawPoints[[i]]]; edgeInsert[n1, n2]; n1 = n2; 
            If[graphicalDebugging === True, graphGraphics[]]];(* 
            second step : tessellation *) 
          leftMostNode = graphSortEdgesAndNodes[]; neighboringNodes = {}; 
          graphMarkOutline[leftMostNode]; graphSimplePolygons[]; 
          If[graphicalDebugging === True, graphGraphics[]]; True, 
          PolygonTessellation]; 
      If[True === catchResult, Map[nodes[[#, 1]] &, polygons, {2}], 
        Message[PolygonTessellation::internal, catchResult]; {}]];

PolygonTessellation3D[points3d_, opts___] := 
    Module[{points2d, plane, planeNormal, normal, basis1, basis2, 
        tessellation}, {planeNormal} = {PlaneNormal} /. {opts} /. 
          Options[PolygonTessellation3D]; 
      If[MatchQ[planeNormal, {_, _, _}], normal = normalize[N[planeNormal]]; 
        basis1 = normalize[First[NullSpace[{normal}]]]; 
        basis2 = normalize[Cross[normal, basis1]]; 
        plane = {Plus @@ N[points3d]/Length[points3d], normal, basis1, 
            basis2}, plane = getPlane[points3d]]; 
      points2d = Map[projectPointToPlane[#, plane] &, points3d]; 
      tessellation = PolygonTessellation[points2d, opts]; 
      Map[embedPointIn3D[#, plane] &, tessellation, {2}]];

TessellatePolygons[expr_, opts___] := 
    Module[{pts, tessellation}, 
      expr /. Polygon[
            pts_List] :> (tessellation = PolygonTessellation[pts, opts]; 
            If[catchResult =!= True, Polygon[pts], 
              Map[Polygon[#] &, tessellation]])];

TessellatePolygons3D[expr_, opts___] := 
    Module[{poly}, expr /. poly_Polygon :> tessellatePolygon3D[poly, opts]];

On[General::spell1];

End[];

EndPackage[];


(*: Examples :*)

(*
<< PolygonTessellation`
*)
(*
points = Table[{Random[Integer, {0, 4}], Random[Integer, {0, 4}]}, {20}];
*)
(*
PolygonTessellation[points, GraphicalDebugging -> True]
*)
(*
points = Table[{Random[Integer, {0, 4}], Random[Integer, {0, 4}], 
        Random[Real, {0, 0.1}]}, {6}];
*)
(*
Show[Graphics3D[Polygon[points]]];
*)
(*
PolygonTessellation3D[points, GraphicalDebugging -> True]
*)
(*
Show[Graphics3D[Map[Polygon, %]]];
*)
(*
poly = Graphics3D[
      Polygon[Table[{Random[Integer, {0, 4}], Random[Integer, {0, 4}], 
            Random[Real, {0, 0.1}]}, {6}]]];
*)
(*
Show[poly];
*)
(*
Show[TessellatePolygons3D[poly, GraphicalDebugging -> True]];
*)
(*
poly = Graphics[
      Polygon[Table[{Random[Integer, {0, 4}], 
            Random[Integer, {0, 4}]}, {6}]]];
*)
(*
Show[poly];
*)
(*
Show[TessellatePolygons[poly, GraphicalDebugging -> False]];
*)
(*
Show[TessellatePolygons[poly, GraphicalDebugging -> False, 
      WindingRule -> EvenQ]];
*)
(*
Show[TessellatePolygons[poly, GraphicalDebugging -> False, 
      WindingRule -> (True &)]];
*)
